/* ui.c -- part of vulcan
 *
 * This program is copyright (C) 2006 Mauro Persano, and is free
 * software which is freely distributable under the terms of the
 * GNU public license, included as the file COPYING in this
 * distribution.  It is NOT public domain software, and any
 * redistribution not permitted by the GNU General Public License is
 * expressly forbidden without prior written permission from
 * the author.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <assert.h>

#include "version.h"
#include "panic.h"
#include "list.h"
#include "dict.h"
#include "render.h"
#include "vector.h"
#include "matrix.h"
#include "mesh.h"
#include "bsptree.h"
#include "model.h"
#include "list.h"
#include "pick.h"
#include "move.h"
#include "game.h"
#include "engine.h"
#include "font.h"
#include "font_render.h"
#include "menu.h"
#include "ui_util.h"
#include "ui_state.h"
#include "ui.h"

#define FONT_FILE "data/font/VeraSe.fontdef"
#define BIG_FONT_FILE "data/font/VeraSe-big.fontdef"

#define FONT_NAME "VeraSe"
#define BIG_FONT_NAME "VeraSe_big"

extern long msecs(void);

struct ui_context ui_context;

enum {
	DBL_CLICK_INTERVAL = 800,	/* in msecs */
	NUM_UI_STATES = 3
};

static union ui_state *ui_states[NUM_UI_STATES] = {
	(union ui_state *)&state_main_menu,
	(union ui_state *)&state_waiting_for_move,
	(union ui_state *)&state_animation,
};

extern void swap_buffers(void);

void
ui_redraw(void)
{
	glClearColor(0.f, 0.f, 0.f, 0.f);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	ui_context.cur_state->common.redraw(ui_context.cur_state);

	swap_buffers();
}

void
ui_reshape(int width, int height)
{
	glViewport(0, 0, width, height);

	ui_context.cur_width = width;
	ui_context.cur_height = height;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45., (GLdouble)width/(GLdouble)height, 1., 1000.);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

static void
init_opengl_state(void)
{
	glClearColor(0.f, 0.f, 0.f, 0.f);
	glClearDepth(1.f);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
}

static void
init_font(void)
{
	struct dict *font_dict;
	struct font_info *fi;

	font_dict = dict_make();

	fontdef_parse_file(font_dict, FONT_FILE);
	fontdef_parse_file(font_dict, BIG_FONT_FILE);

	if ((fi = dict_get(font_dict, FONT_NAME)) == NULL)
		panic("font %s found in fontdef file?", FONT_NAME);

	ui_context.font = font_render_info_make(fi);

	if ((fi = dict_get(font_dict, BIG_FONT_NAME)) == NULL)
		panic("font %s not found in fontdef file?", BIG_FONT_NAME);

	ui_context.font_big = font_render_info_make(fi);
}

static void
init_ui_states(void)
{
	int i;

	for (i = 0; i < NUM_UI_STATES; i++)
		ui_states[i]->common.initialize(ui_states[i]);
}

void
gl_initialize(void)
{
	mat_make_rotation_around_z(&ui_context.rmatrix, -M_PI/2.);
	vec_set(&ui_context.board_pos, 0.f, 0.f, -400.f);
	init_opengl_state();
	init_render();
	init_font();
	init_ui_states();
}

void
ui_set_state(enum ui_state_type state_type)
{
	assert(state_type >= 0 && state_type < NUM_UI_STATES);

	ui_context.cur_state = ui_states[state_type];

	ui_context.cur_state->common.on_state_switch(ui_context.cur_state);
}

void
ui_initialize(enum player_type white_player, enum player_type black_player,
  int max_depth)
{
	init_move_tables();
	init_engine();

	init_board_state(&ui_context.state);

	memset(&ui_context.selected_squares, UNSELECTED_SQUARE,
	  sizeof ui_context.selected_squares);

	ui_context.computer_player_state = IDLE;
	ui_context.cur_side = 0;
	ui_context.left_button_down = 0;
	ui_context.right_button_down = 0;
	ui_context.control_key_down = 0;
	ui_context.start_position_selected = 0;
	ui_context.pos_highlighted = 0;
	ui_context.last_clicked_x = -1;
	ui_context.last_clicked_y = -1;
	ui_context.last_click_msecs = 0;
	ui_context.max_depth = max_depth;

	ui_context.players[0] = white_player;
	ui_context.players[1] = black_player;

	ui_context.last_move_buf[0] = 0;

	ui_context.cur_width = -1;
	ui_context.cur_height = -1;

	ui_context.to_quit = 0;

	ui_set_state(STATE_MAIN_MENU);
}

void
ui_on_worker_reply(union move *next_move)
{
	ui_context.cur_state->common.on_worker_reply(ui_context.cur_state,
	  next_move);
}

void
ui_on_button_press(int button, int x, int y)
{		
	unsigned long cur_msecs;

	if (button == BTN_LEFT || button == BTN_RIGHT) {
		if (button == BTN_LEFT)
			ui_context.left_button_down = 1;
		else if (button == BTN_RIGHT)
			ui_context.right_button_down = 1;

		cur_msecs = msecs();

/*
		if (button == BTN_LEFT && x == ui_context.last_clicked_x &&
		  y == ui_context.last_clicked_y &&
		  cur_msecs - ui_context.last_click_msecs <= DBL_CLICK_INTERVAL) {
			ui_context.last_click_msecs = 0;
			ui_context.last_clicked_x = ui_context.last_clicked_y = -1;
			ui_context.cur_state->common.on_double_click(
			  ui_context.cur_state, button, x, y);
		} else 
*/
		{
			ui_context.last_click_msecs = cur_msecs;
			ui_context.last_clicked_x = x;
			ui_context.last_clicked_y = y;

			ui_context.cur_state->common.on_button_press(
			  ui_context.cur_state, button, x, y);
		}
	}
}

void
ui_on_button_release(int button, int x, int y)
{
	if (button == BTN_LEFT)
		ui_context.left_button_down = 0;
	else if (button == BTN_RIGHT)
		ui_context.right_button_down = 0;

	ui_context.cur_state->common.on_button_release(ui_context.cur_state,
	  button, x, y);
}

void
ui_on_motion_notify(int button, int x, int y)
{
	if (ui_context.left_button_down || ui_context.right_button_down) {
		ui_context.cur_state->common.on_drag(ui_context.cur_state,
		  ui_context.left_button_down ? BTN_LEFT : BTN_RIGHT, x, y);
	} else {
		ui_context.cur_state->common.on_motion_notify(
		  ui_context.cur_state, button, x, y);
	}
}

void
ui_on_idle(void)
{
	ui_context.cur_state->common.on_idle(ui_context.cur_state);
}
