#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "panic.h"
#include "vector.h"
#include "matrix.h"
#include "pick.h"
#include "game.h"
#include "move.h"
#include "font.h"
#include "font_render.h"
#include "ui.h"
#include "ui_util.h"
#include "ui_state.h"

extern int send_worker_request(void);

static void
game_do_move(union move *move)
{
	sprintf(ui_context.last_move_buf, "%d. %s %s",
	  1 + ui_context.state.cur_ply/2,
	  (ui_context.state.cur_ply & 1) ? "..." : "",
	  move_as_string(&ui_context.state, move));

	do_move(&ui_context.state, move, &ui_context.last_undo_info);

	if (get_game_status(&ui_context.state) != IN_PLAY) {
		ui_context.computer_player_state =
		  ui_context.cur_side == BLACK_FLAG ?
		    DONE_MATE_FOR_BLACK : DONE_MATE_FOR_WHITE;

		strcat(ui_context.last_move_buf, "++");
	} else {
		ui_context.cur_side ^= BLACK_FLAG;
		ui_context.computer_player_state = IDLE;

		if (is_in_check(&ui_context.state, ui_context.cur_side))
			strcat(ui_context.last_move_buf, "+");
	}

	if (move->type == PIECE_MOVE) {
		ui_context.animation_move = *move;
		ui_set_state(STATE_ANIMATION);
	}
}

static enum player_type
cur_player_type(void)
{
	return ui_context.players[ui_context.cur_side == BLACK_FLAG];
}

static void
on_idle(union ui_state *state)
{
	if (ui_context.computer_player_state == IDLE) {
		if (cur_player_type() == COMPUTER_PLAYER) {
			if (send_worker_request() != 0)
				panic("couldn't send worker request?");
			ui_context.computer_player_state = THINKING;
		} else {
			ui_context.computer_player_state = WAITING_FOR_USER_INPUT;
		}
	}
}

static void
mark_selected_square(struct selected_squares *selected,
  const struct position *pos, enum selection_type t)
{
	selected->selection[pos->level][pos->square] = t;
}

static void
mark_selected_attack_board(struct selected_squares *selected,
  const struct attack_board *pos, enum selection_type t)
{
	int i, j;
	struct position first_pos;
	unsigned char *p;

	get_attack_board_position(&first_pos, pos->main_board, pos->position);

	p = &selected->selection[first_pos.level][first_pos.square];

	for (i = 0; i < ATTACK_BOARD_SIZE; i++) {
		for (j = 0; j < ATTACK_BOARD_SIZE; j++)
			*p++ = t;

		p += BCOLS - ATTACK_BOARD_SIZE;
	}
}

static void
mark_selected_squares(struct selected_squares *selected,
  const struct board_state *state, 
  int side, union move *moves, int num_moves)
{
	union move *move, *end;
	const struct position *pos;
	unsigned char s;
	enum selection_type t;
	struct position closest;

	end = &moves[num_moves];

	for (move = moves; move != end; move++) {
		switch (move->type) {
			case PIECE_MOVE:
				mark_selected_square(selected,
				  &move->piece_move.from, SELECTED_SQUARE);

				pos = &move->piece_move.to;

				s = state->board[pos->level][pos->square];

				if (s == EMPTY || (s & BLACK_FLAG) == side)
					t = SELECTED_SQUARE;
				else
					t = ATTACKED_SQUARE;

				mark_selected_square(selected, pos, t);
				break;

			case ATTACK_BOARD_MOVE:
				mark_selected_attack_board(selected,
				  &move->attack_board_move.from,
				    SELECTED_SQUARE);

				/* mark clickable destination square */

				get_closest_main_board_square_for_attack_board(
				  &closest, &move->attack_board_move.to);

				mark_selected_square(selected, &closest,
				  ATTACKED_SQUARE);
				break;

			case WHITE_KINGSIDE_CASTLING:
				selected->selection[4][11*BCOLS + 5] =
				  SELECTED_SQUARE;
				selected->selection[4][11*BCOLS + 6] =
				  ATTACKED_SQUARE;
				break;

			case WHITE_QUEENSIDE_CASTLING:
				selected->selection[4][11*BCOLS + 5] =
				  SELECTED_SQUARE;
				selected->selection[4][11*BCOLS + 2] =
				  SELECTED_SQUARE;
				break;

			case BLACK_KINGSIDE_CASTLING:
				selected->selection[0][2*BCOLS + 5] =
				  SELECTED_SQUARE;
				selected->selection[0][2*BCOLS + 6] =
				  ATTACKED_SQUARE;
				break;

			case BLACK_QUEENSIDE_CASTLING:
				selected->selection[0][2*BCOLS + 5] =
				  SELECTED_SQUARE;
				selected->selection[0][2*BCOLS + 2] =
				  SELECTED_SQUARE;
				break;

			default:
				assert(0);
		}
	}
}

static void
on_start_position_selected(struct position *pos)
{
	union move moves[MAX_MOVES];
	int n;

	if ((n = get_legal_moves_for_position(moves, pos, &ui_context.state,
	  ui_context.cur_side)) > 0) {
		mark_selected_squares(&ui_context.selected_squares,
		  &ui_context.state, ui_context.cur_side, moves, n);

		mark_selected_square(&ui_context.selected_squares, pos,
		  SELECTED_SQUARE);

		ui_context.last_selected_pos = *pos;
		ui_context.start_position_selected = 1;
		ui_context.pos_highlighted = 0;
	}
}

static void
on_end_position_selected(struct position *pos)
{
	union move moves[MAX_MOVES];
	union move *selected_move;
	struct position closest, *start;
	int i, n;

	start = &ui_context.last_selected_pos;

	assert((n = get_legal_moves_for_position(moves,
	  start, &ui_context.state, ui_context.cur_side)) > 0);

	selected_move = NULL;

	for (i = 0; i < n && selected_move == NULL; i++) {
		switch (moves[i].type) {
			case PIECE_MOVE:
				if (SAME_POS(&moves[i].piece_move.to, pos)) {
					selected_move = &moves[i];
					break;
				}
				break;

			case ATTACK_BOARD_MOVE:
				get_closest_main_board_square_for_attack_board(
				  &closest, &moves[i].attack_board_move.to);

				if (SAME_POS(&closest, pos)) {
					selected_move = &moves[i];
					break;
				}
				break;

			case WHITE_KINGSIDE_CASTLING:
				if (start->level == 4 &&
				  start->square == 11*BCOLS+5 &&
				  pos->level == 4 &&
				  pos->square == 11*BCOLS+6) {
					selected_move = &moves[i];
					break;
				}
				break;

			case WHITE_QUEENSIDE_CASTLING:
				if (start->level == 4 &&
				  start->square == 11*BCOLS+5 &&
				  pos->level == 4 &&
				  pos->square == 11*BCOLS+2) {
					selected_move = &moves[i];
					break;
				}
				break;

			case BLACK_KINGSIDE_CASTLING:
				if (start->level == 0 &&
				  start->square == 2*BCOLS+5 &&
				  pos->level == 0 &&
				  pos->square == 2*BCOLS+6) {
					selected_move = &moves[i];
					break;
				}
				break;

			case BLACK_QUEENSIDE_CASTLING:
				if (start->level == 0 &&
				  start->square == 2*BCOLS+5 &&
				  pos->level == 0 &&
				  pos->square == 2*BCOLS+2) {
					selected_move = &moves[i];
					break;
				}
				break;
		}
	}

	if (selected_move != NULL)
		game_do_move(selected_move);

	ui_context.start_position_selected = 0;
}

static void
on_position_selected(struct position *pos)
{
	memset(&ui_context.selected_squares, UNSELECTED_SQUARE,
	  sizeof ui_context.selected_squares);

	if (!ui_context.start_position_selected)
		on_start_position_selected(pos);
	else
		on_end_position_selected(pos);
}

/* state handlers */

static void
on_worker_reply(union ui_state *state, union move *move)
{
	assert(ui_context.computer_player_state == THINKING);
	game_do_move(move);
}

static void
on_drag(union ui_state *state, int button, int x, int y)
{
	int dx, dy;
	struct matrix r, rx, ry;

	if (ui_context.pos_highlighted) {
		mark_selected_square(&ui_context.selected_squares,
		  &ui_context.last_highlighted_pos, UNSELECTED_SQUARE);

		ui_context.pos_highlighted = 0;
	}

	dx = x - ui_context.last_clicked_x;
	dy = y - ui_context.last_clicked_y;
	ui_context.last_clicked_x = x;
	ui_context.last_clicked_y = y;

	if (button == BTN_LEFT) {
		mat_make_rotation_around_y(&ry, (float)-dx/100.f);
		mat_make_rotation_around_x(&rx, (float)dy/100.f);
		mat_mul_copy(&r, &ry, &rx);
		mat_mul_copy(&ui_context.rmatrix, &r, &ui_context.rmatrix);
	} else if (ui_context.control_key_down) {
		ui_context.board_pos.z -= (float)dy/1.5f;
	} else {
		ui_context.board_pos.x -= (float)dx*ui_context.board_pos.z/500.f;
		ui_context.board_pos.y += (float)dy*ui_context.board_pos.z/500.f;
	}
}

static void
on_motion_notify(union ui_state *state, int button, int x, int y)
{
	if (ui_context.computer_player_state == WAITING_FOR_USER_INPUT) {
		struct position pos;
		struct matrix mv;

		if (ui_context.pos_highlighted) {
			mark_selected_square(&ui_context.selected_squares,
			  &ui_context.last_highlighted_pos, UNSELECTED_SQUARE);

			ui_context.pos_highlighted = 0;
		}

		ui_set_modelview_matrix(&mv);

		if (pick_square(&pos, &ui_context.state, &mv, x, y)) {
			const int s = ui_context.state.board[pos.level][pos.square];

			if (s != EMPTY && (s & BLACK_FLAG) == ui_context.cur_side) {
				if (!(ui_context.start_position_selected && ui_context.selected_squares.selection[pos.level][pos.square] == SELECTED_SQUARE)) {
					ui_context.pos_highlighted = 1;
					ui_context.last_highlighted_pos = pos;

					mark_selected_square(
					  &ui_context.selected_squares,
					  &ui_context.last_highlighted_pos,
					  HIGHLIGHTED_PIECE);
				}
			}
		}
	}
}

static int
is_same_position(struct position *a, struct position *b)
{
	return a->level == b->level && a->square == b->square;
}

static void
on_button_press(union ui_state *state, int button, int x, int y)
{
	struct position pos;
	struct matrix mv;

	if (ui_context.computer_player_state == WAITING_FOR_USER_INPUT
	  && button == BTN_LEFT) {
		ui_set_modelview_matrix(&mv);

		if (pick_square(&pos, &ui_context.state, &mv, x, y)) {
			if (!ui_context.start_position_selected ||
			  !is_same_position(&ui_context.last_selected_pos,
			    &pos)) {
				memset(&ui_context.selected_squares,
				  UNSELECTED_SQUARE,
				  sizeof ui_context.selected_squares);

				on_position_selected(&pos);
			}
		}
	}
}

static void
on_double_click(union ui_state *state, int button, int x, int y)
{
	on_button_press(state, button, x, y);
}

static void
on_button_release(union ui_state *state, int button, int x, int y)
{
}

static void
render_text(void)
{
	int char_height;
	const int baseline = 22;

	glColor4f(1.f, 1.f, 1.f, 1.f);

	char_height = ui_context.font->char_height;

	render_string(ui_context.font, ui_context.last_move_buf, 5,
	  ui_context.cur_height - 22);

	switch (ui_context.computer_player_state) {
		case DONE_MATE_FOR_WHITE:
			render_string(ui_context.font, "Mate for white!",
			  5, ui_context.cur_height - (baseline + char_height + 2));
			break;

		case DONE_MATE_FOR_BLACK:
			render_string(ui_context.font, "Mate for black!",
			  5, ui_context.cur_height - (baseline + char_height + 2));
			break;

		default:
			render_string(ui_context.font,
			  ui_context.cur_side == BLACK_FLAG ?
			  "Black's turn." : "White's turn.",
			  5, ui_context.cur_height - (baseline + char_height + 2));
			render_string(ui_context.font,
  			  cur_player_type() == COMPUTER_PLAYER ?
			    "Computer thinking..." : "Your move?",
			  5, ui_context.cur_height - (baseline + 2*(char_height + 2)));
			break;
	}
}

static void
redraw(union ui_state *state)
{
	ui_render_board_state();
	render_text();
}

static void
initialize(union ui_state *state)
{
}

static void
on_state_switch(union ui_state *state)
{
}

struct state_waiting_for_move state_waiting_for_move = {
	{
		on_idle,
		on_worker_reply,
		on_motion_notify,
		on_drag,
		on_button_press,
		on_double_click,
		on_button_release,
		redraw,
		initialize,
		on_state_switch,
	}
};
